from typing import Optional, List

import vtk
import vmi
import numpy as np
import numpy.random.common
import numpy.random.bounded_integers
import numpy.random.entropy

vmi.CS4x4 = None


class CS4x4:
    """
    坐标系，以4x4矩阵表示的局部坐标系或坐标变换
    """

    def __init__(self, axis0: Optional[np.ndarray] = None, axis1: Optional[np.ndarray] = None,
                 axis2: Optional[np.ndarray] = None, origin: Optional[np.ndarray] = None):
        """
        局部坐标系

        :param axis0: 3x1第一坐标轴，转为齐次坐标构成矩阵的第一行
        :param axis1: 3x1第二坐标轴，转为齐次坐标构成矩阵的第二行
        :param axis2: 3x1第三坐标轴，转为齐次坐标构成矩阵的第三行
        :param origin: 3x1坐标原点，转为齐次坐标构成矩阵的第四行
        """
        self._axis3x3 = np.identity(3)
        self._origin3 = np.zeros(3)
        if origin is not None:
            self.origin(origin)
        if axis0 is not None:
            self.axis(0, axis0)
        if axis1 is not None:
            self.axis(1, axis1)
        if axis2 is not None:
            self.axis(2, axis2)

    def __setstate__(self, s):
        self.__init__()
        self.__dict__.update(s)
        s = self.__dict__

    def __getstate__(self):
        s = self.__dict__.copy()
        for kw in []:
            if kw in s:
                del s[kw]
        return s

    def __repr__(self):
        return self.ary4x4().__repr__()

    def copy(self) -> vmi.CS4x4:
        """
        复制实例

        :return: vmi.CS4x4 - 具有相同矩阵的实例
        """
        cs = CS4x4()
        cs._axis3x3 = self._axis3x3.copy()
        cs._origin3 = self._origin3.copy()
        return cs

    def ary4x4(self, ary: Optional[np.ndarray] = None) -> np.ndarray:
        """
        返回当前矩阵，或设置新矩阵

        :param ary: 4x4新矩阵
        :return: np.ndarray - 当前矩阵
        """
        if ary is not None:
            self._origin3 = ary[3][:3]
        return np.hstack((np.vstack((self._axis3x3, self._origin3)), ([0], [0], [0], [1])))

    def ary4x4inv(self) -> np.ndarray:
        """返回当前矩阵的逆矩阵"""
        return np.linalg.inv(self.ary4x4())

    def vtkmat4x4(self, mat: vtk.vtkMatrix4x4 = None) -> vtk.vtkMatrix4x4:
        """返回当前矩阵的vtk矩阵，或设置新vtk矩阵"""
        if mat is not None:
            ary = np.zeros(4, 4)
            for j in range(4):
                for i in range(4):
                    ary[i, j] = mat.GetElement(j, i)

        ary = self.ary4x4()
        matrix = vtk.vtkMatrix4x4()
        for j in range(4):
            for i in range(4):
                matrix.SetElement(j, i, ary[i, j])
        return matrix

    def mat4x4inv(self) -> vtk.vtkMatrix4x4:
        """返回当前矩阵的vtkMatrix4x4逆矩阵"""
        ary = self.ary4x4inv()
        matrix = vtk.vtkMatrix4x4()
        for j in range(4):
            for i in range(4):
                matrix.SetElement(j, i, ary[i, j])
        return matrix

    def inv(self) -> vmi.CS4x4:
        """返回逆变换"""
        cs = CS4x4()
        cs.ary4x4(self.ary4x4inv())
        return cs

    def axis(self, i: int, arg: Optional[np.ndarray] = None) -> np.ndarray:
        """
        返回第i坐标轴，或设置新坐标轴
        :param i: 坐标轴次序
        :param arg: 3x1新坐标轴
        :return: 3x1当前第i坐标轴
        """
        if arg is not None:
            t = np.array(self._axis3x3)
            t[i] = np.array(arg[:3])
            self._axis3x3[i] = np.array(arg)

        return self._axis3x3[i]

    def origin(self, arg: Optional[np.ndarray] = None) -> np.ndarray:
        """
        返回坐标原点

        :param arg: 3x1新坐标原点
        :return: 3x1坐标原点
        """
        if arg is not None:
            self._origin3 = np.array(arg[:3])
        return self._origin3

    def orthogonalize(self, i_list: List[int]) -> None:
        """
        按照次序列表正交化坐标轴，例如i_list=[0, 2]表示先用z轴y轴正交化x轴，再用x轴y轴正交化z轴

        :param i_list: 坐标轴次序列表
        :return: None
        """
        for i in i_list:
            self.axis(i, np.cross(self.axis((i + 1) % 3), self.axis((i + 2) % 3)))

    def normalize(self, i_list: List[int]) -> None:
        """
        按照次序列表单位化坐标轴，例如i_list=[0, 2]表示将x轴z轴单位化

        :param i_list: 坐标轴次序列表
        :return: None
        """
        if i_list is None:
            i_list = [0, 1, 2]
        for i in i_list:
            self.axis(i, self.axis(i) / np.linalg.norm(self.axis(i)))

    def mpt(self, pt: np.ndarray, origin: Optional[np.ndarray] = None,
            other: Optional[vmi.CS4x4] = None) -> np.ndarray:
        """
        坐标系变换：点pt由当前坐标系变换到世界坐标系，如果设置了other则变换到other坐标系
        坐标变换：世界坐标系中的点pt经过当前变换得到新点，如果设置了origin则在此次计算中原点暂时设为origin

        :param pt: 3x1点坐标
        :param origin: 3x1坐标原点
        :param other: 坐标系
        :return: 变换后坐标
        """
        origin_ = self.origin()
        if origin is not None:
            self.origin(origin)

        pt = np.append(np.array(pt[:3]), 1)
        pt = pt @ self.ary4x4()
        if other is not None:
            pt = pt @ other.ary4x4inv()

        if origin is not None:
            self.origin(origin_)
        return pt[:3]

    def mvt(self, vt: Optional[np.ndarray], other: Optional[vmi.CS4x4] = None) -> np.ndarray:
        """
        坐标系变换：向量vt由当前坐标系变换到世界坐标系，如果设置了other则变换到other坐标系
        坐标变换：世界坐标系中的向量vt经过当前变换，获得新向量

        :param vt: 3x1向量
        :param other: 坐标系
        :return: 变换后坐标
        """
        vt = np.append(np.array(vt[:3]), 0)
        vt = vt @ self.ary4x4()
        if other is not None:
            vt = vt @ other.ary4x4inv()
        return vt[:3]

    def mcs(self, cs: vmi.CS4x4):
        """
        坐标系变换：坐标系cs由当前坐标系变换到世界坐标系，如果设置了other则变换到other坐标系
        坐标变换：世界坐标系中的局部坐标系cs经过当前变换得到新坐标系
        """
        cs = cs.ary4x4() @ self.ary4x4()
        return cs

    def translate(self, vt: np.ndarray) -> vmi.CS4x4:
        """
        串接平移变换
        :param vt: 3x1平移变换
        :return: 新坐标系
        """
        mat = np.array([[1, 0, 0, 0],
                        [0, 1, 0, 0],
                        [0, 0, 1, 0],
                        [vt[0], vt[1], vt[2], 1]])
        self.ary4x4(self.ary4x4() @ mat)
        return self

    def rotate(self, angle: float, axis: np.ndarray, origin: np.ndarray = None) -> vmi.CS4x4:
        """
        串接旋转变换
        https://en.wikipedia.org/wiki/Rotation_matrix#In_three_dimensions

        :param angle: 旋转角度（deg）
        :param axis: 3x1旋转轴向量
        :param origin: 3x1旋转轴经过的一点
        :return: 新坐标系
        """
        axis = np.array(axis[:3])
        if np.linalg.norm(axis) == 0:
            raise np.linalg.LinAlgError('旋转轴向错误')

        angle = np.radians(angle)

        mat = np.cos(angle) * np.identity(3) + np.sin(angle) * np.array(
            [[0, axis[2], -axis[1]],
             [-axis[2], 0, axis[0]],
             [axis[1], -axis[0], 0]]) + (1 - np.cos(angle)) * np.outer(axis, axis)
        mat = np.hstack((np.vstack((mat, [0, 0, 0])), ([0], [0], [0], [1])))

        if origin is None:
            self.ary4x4(self.ary4x4() @ mat)
        else:
            origin = np.array(origin[:3])
            self.translate(-origin)
            self.ary4x4(self.ary4x4() @ mat)
            self.translate(origin)
        return self

    def reflect(self, normal: np.ndarray, origin: Optional[np.ndarray] = None) -> vmi.CS4x4:
        """
        串接反射变换
        https://en.wikipedia.org/wiki/Reflection_(mathematics)#Reflection_through_a_hyperplane_in_n_dimensions

        :param normal: 反射面法向量
        :param origin: 反射面经过的一点
        :return: 新坐标系
        """
        normal = np.array(normal[:3])
        if np.linalg.norm(normal) == 0:
            raise np.linalg.LinAlgError('反射法向错误')

        mat = np.identity(3) - 2 / np.linalg.norm(normal) ** 2 * np.outer(normal, normal)
        mat = np.hstack((np.vstack((mat, [0, 0, 0])), ([0], [0], [0], [1])))

        if origin is None:
            self.ary4x4(self.ary4x4() @ mat)
        else:
            origin = np.array(origin[:3])
            self.translate(-origin)
            self.ary4x4(self.ary4x4() @ mat)
            self.translate(origin)
        return self

    def scale(self, factor: float, origin: Optional[np.ndarray] = None) -> vmi.CS4x4:
        """
        串接缩放变换

        :param factor: 缩放比例
        :param origin: 缩放投影中心
        :return: 新坐标系
        """
        mat = np.identity(3) * np.array(factor[:3])
        mat = np.hstack((np.vstack((mat, [0, 0, 0])), ([0], [0], [0], [1])))

        if origin is None:
            self.ary4x4(self.ary4x4() @ mat)
        else:
            origin = np.array(origin[:3])
            self.translate(-origin)
            self.ary4x4(self.ary4x4() @ mat)
            self.translate(origin)
        return self


def CS4x4_InterLandmarks(pts0: List[np.ndarray], pts1: List[np.ndarray]) -> vmi.CS4x4:
    """
    对应点列变换，计算点列pts0到点列pts1的最小二乘变换

    :param pts0: 点列
    :param pts1: 点列
    :return: 坐标系
    """
    if len(pts0) > 2 and len(pts1) > 2:
        pts = [vtk.vtkPoints(), vtk.vtkPoints()]
        for pt in pts0:
            pts[0].InsertNextPoint([pt[0], pt[1], pt[2]])
        for pt in pts1:
            pts[1].InsertNextPoint([pt[0], pt[1], pt[2]])

        lt = vtk.vtkLandmarkTransform()
        lt.SetSourceLandmarks(pts[0])
        lt.SetTargetLandmarks(pts[1])
        lt.SetModeToRigidBody()
        lt.Update()

        mat = vtk.vtkMatrix4x4()
        lt.GetMatrix(mat)

        cs = CS4x4()
        cs.vtkmat4x4(mat)
        return cs


def ptOnPlane(pt: np.ndarray, origin: np.ndarray, normal: np.ndarray) -> np.ndarray:
    """
    计算点到平面投影

    :param pt: 初始点
    :param origin: 相关平面经过一点
    :param normal: 相关平面法向量
    :return: 投影后的点
    """
    pt, origin, normal = np.array(pt), np.array(origin), np.array(normal)
    return pt - normal * np.dot(pt - origin, normal)


def vtOnPlane(vt: np.ndarray, normal: np.ndarray) -> np.ndarray:
    """
    计算向量到平面投影

    :param vt: 初始向量
    :param normal: 相关平面法向量
    :return: 投影后的向量
    """
    vt, normal = np.array(vt), np.array(normal)
    n2 = np.dot(normal, normal)
    n2 = n2 if n2 != 0 else 1
    t = np.dot(vt, normal)
    return vt - t * normal / n2


def vtOnVector(vt: np.ndarray, vector: np.ndarray) -> np.ndarray:
    """
    计算向量到向量投影
    :param vt: 初始向量
    :param vector: 相关向量
    :return: 投影后的向量
    """
    dot = np.dot(vector, vector)
    if dot == 0:
        return np.zeros(3)
    return np.dot(vt, vector) / dot * vector



import setuptools

name = 'vmi'
version = '0.3.3'

with open('README.md', 'r', encoding='UTF-8') as fh:
    long_description = fh.read()

setuptools.setup(
    name=name,
    version=version,
    author='sjtu_6547',
    author_email='88172828@qq.com',
    description='Visual Medical Innovation powered by WANGYiping',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://pypi.org/project/vmi/',
    keywords='Qt PySide2 SimpleITK itk vtk pythonocc-core OpenCASCADE CGAL',
    packages=setuptools.find_packages(),
    python_requires='>=3.6',
    install_requires=['PySide2>=5.13',
                      'SimpleITK>=1.2',
                      'vtk>=8.1',
                      'pydicom>=1.3',
                      'numpy'],
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Win32 (MS Windows)',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.7',
        'Topic :: Multimedia :: Graphics :: 3D Modeling',
        'Topic :: Multimedia :: Graphics :: 3D Rendering',
        'Topic :: Scientific/Engineering :: Medical Science Apps.',
        'Topic :: Scientific/Engineering :: Visualization',
    ],
)



from typing import Optional, List

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *
from PySide2.QtWinExtras import *
import threading
import time
import os
import sys
import pathlib
import shutil

app = QApplication()
app.setOrganizationName('vmi')
app.setApplicationName('vmi')
app.setStyle(QStyleFactory.create('Fusion'))

appFonts = {}

desktopPath: str = QStandardPaths.writableLocation(QStandardPaths.DesktopLocation)
appDataPath: str = QStandardPaths.writableLocation(QStandardPaths.AppDataLocation)
os.makedirs(appDataPath, exist_ok=True)

_AppWindow: QWidget = None
_TaskbarButton = QWinTaskbarButton()
_Progress = _TaskbarButton.progress()


def _appinit() -> None:
    fonts = {'alizh': ':/qrc/font/Alibaba-PuHuiTi-Regular.otf',
             'alien': ':/qrc/font/AlibabaSans-Regular.otf'}

    for kw in fonts:
        f = QFile(fonts[kw])
        f.open(QIODevice.ReadOnly)
        bts = f.readAll()
        f.close()

        p = pathlib.Path(appDataPath) / kw
        p.write_bytes(bts)
        appFonts[kw] = str(p)


_appinit()


def appexit() -> None:
    """
    退出主程序

    :return: None
    """
    shutil.rmtree(appDataPath, ignore_errors=True)
    sys.exit()


def appwindow(w: Optional[QWidget] = None) -> QWidget:
    """
    设置主程序窗口

    :param w: 主程序窗口
    :return: QWidget - 返回当前主程序窗口
    """
    if w is not None:
        global _AppWindow
        _AppWindow = w
        _TaskbarButton.setParent(w)
        _TaskbarButton.setWindow(w.windowHandle())
        _Progress.setVisible(True)
        _Progress.setRange(0, 100)
        _Progress.setValue(0)
    return _AppWindow


def appexec(w: QWidget, width: int = 1600, height: int = 900) -> int:
    """
    设置主程序窗口，居中显示，启动主程序

    :param w: 主程序窗口
    :param width: 窗口宽度
    :param height: 窗口高度
    :return: int - 主程序结束状态
    """
    appwindow(w)
    rect = QGuiApplication.screens()[0].geometry()
    rect = QRect(rect.center() - QPoint(int(width / 2), int(height / 2)), QSize(width, height))
    w.setGeometry(rect)
    w.show()
    app.exec_()


def appwait(thread: threading.Thread, progress: Optional[List[int]] = None) -> None:
    """
    执行后台线程并等待，默认在windows系统任务栏循环更新进度，设置progress=[value]可以显示真实进度

    :param thread: 后台线程
    :param progress: 进度
    :return: None
    """
    while thread.isAlive():
        if progress is None:
            value = (_Progress.value() + 5) % 100
        else:
            value = progress[0]
        _Progress.setValue(value)
        time.sleep(0.1)
    _Progress.setValue(0)


_scripts = pathlib.Path(sys.executable).parent / 'Scripts'
_lupdate = str(_scripts / 'pyside2-lupdate')
_uic = str(_scripts / 'pyside2-uic')
_rcc = str(_scripts / 'pyside2-rcc')

_pyside2 = pathlib.Path(sys.executable).parent / 'Lib/site-packages/PySide2'
_designer = str(_pyside2 / 'designer')
_linguist = str(_pyside2 / 'linguist')
_lrelease = str(_pyside2 / 'lrelease')


def app_uic(ui_dir: str = 'ui', py_dir: str = '.', prefix: str = 'Ui_') -> None:
    """
    将ui_dir目录下的所有.ui文件转换到py_dir目录下.py文件并附加前缀ui_

    :param ui_dir: 界面文件（.ui）目录
    :param py_dir: 转换后的源代码文件（.py）目录
    :param prefix: 转换后的文件名前缀
    :return: None
    """
    py_dir = pathlib.Path(py_dir)

    for ui in pathlib.Path(ui_dir).rglob('*.ui'):
        if ui.is_file():
            py = py_dir / (prefix + ui.with_suffix('.py').name)
            os.system(_uic + ' ' + str(ui) + ' -o ' + str(py) + ' -x')
            print('[uic.exe]', _uic, ui, '-o', py, '-x')


def app_lupdate(py_dir: str = '.', ts_file: str = 'zh_CN.ts') -> None:
    """
    将py_dir目录下的所有.py文件更新翻译到同目录下的.ts文件ts_file

    :param py_dir: 源代码文件（.py)目录
    :param ts_file: 转换后的翻译文件（.ts）路径
    :return: None
    """
    py_dir = pathlib.Path(py_dir)
    pys = str()

    for py in pathlib.Path(py_dir).rglob('*.py'):
        if py.is_file():
            pys += str(py) + ' '

    if len(pys) > 0:
        ts_file = py_dir / ts_file
        os.system(_lupdate + ' ' + str(pys) + ' -ts ' + str(ts_file))
        print('[lupdate.exe]', _lupdate, pys, '-ts', ts_file)


def app_lrelease(ts_file: str = './zh_CN.ts',
                 qm_file: str = 'qrc/tr/zh_CN.qm') -> None:
    """
    将可读翻译文件ts_file转换为二进制翻译文件qm_file

    :param ts_file: 可读翻译文件（.ts）
    :param qm_file: 转换后的二进制翻译文件（.qm）
    :return: None
    """
    os.system(_lrelease + ' ' + ts_file + ' -qm ' + qm_file)
    print('[lrelease.exe]', _lrelease, ts_file, '-qm', qm_file)
    os.system(_linguist + ' ' + './zh_CN.ts')


def app_rcc(rc_dir: str = 'qrc', qrc_file: str = 'vmi.qrc', py_file: str = 'vrc.py') -> None:
    """
    将rc_dir目录下的所有文件转换为.qrc资源文件qrc_file，再转换为.py文件py_file

    :param rc_dir: 资源文件目录
    :param qrc_file: 资源文件（.qrc）
    :param py_file: 转换后的源代码文件（.py）
    :return: None
    """
    rc_dir = pathlib.Path(rc_dir)
    indent = ' ' * 4
    rcs = str()

    for rc in pathlib.Path(rc_dir).rglob('*'):
        if rc.is_file():
            rc = str(rc).replace('\\', '/')
            rcs += indent * 2 + '<file>' + rc + '</file>' + '\n'

    if len(rcs) > 0:
        t = '<!DOCTYPE RCC>'
        t += '<RCC version="1.0">' + '\n'
        t += indent + '<qresource prefix="/">' + '\n'
        t += rcs
        t += indent + '</qresource>' + '\n'
        t += '</RCC>'

        with open(qrc_file, 'w') as f:
            f.write(t)

        os.system(_rcc + ' ' + qrc_file + ' -o ' + py_file)
        print('[rcc.exe]', _rcc, qrc_file, '-o', py_file)


if __name__ == '__main__':
    app_uic()  # 更新界面.ui文件
    app_lupdate()  # 更新翻译.ts文件
    app_lrelease()  # 更新翻译.qm文件
    app_rcc()  # 更新.qrc文件

    os.startfile(_designer)  # 打开界面.ui文件



import vmi
import numpy as np


def read_stl():
    """读取*.stl文件"""
    file_name = vmi.askOpenFile('*.stl')
    if file_name is not None:
        pd = vmi.vtkread_STL(file_name)
        if pd.GetNumberOfCells() > 0:
            return pd


def LeftButtonPress(**kwargs):
    """响应左键按下"""
    if kwargs['picked'] is mesh_prop:  # 判断左键按下时是否在模型上
        pt = view.mouseOnPropCell()  # 获得拾取到的世界坐标点
        path_pts.clear()  # 清空拾取路径点列，开始拾取
        path_pts.append(pt)  # 添加第一个拾取点
        update_path(path_pts)  # 更新拾取路径显示 8


def LeftButtonPressMove(**kwargs):
    """响应左键按下之后移动"""
    if kwargs['picked'] is mesh_prop:  # 左键按下移动时，参数传递的仍然是按下时的对象
        pt = view.mouseOnPropCell()  # 获得拾取到的世界坐标点

        for path_pt in path_pts:  # 遍历拾取路径点
            if (path_pt == pt).all():  # 如果已经存在相同的点，则返回
                return

        update_path([*path_pts, pt])  # 更新拾取路径显示，暂不添加拾取路径点

        if view.mouseOnProp() is mesh_prop:  # 判断鼠标当前位置是否在模型上
            path_pts.append(pt)  # 添加拾取路径点


def LeftButtonPressMoveRelease(**kwargs):
    """响应左键按下移动之后释放"""
    if kwargs['picked'] is mesh_prop:  # 左键按下移动释放时，参数传递的仍然是按下时的对象
        update_plate()  # 左键释放时更新导板


def Wheel(**kwargs):
    """响应滚轮"""
    mesh_prop.visibleToggle()  # 切换模型显示与隐藏


def update_path(pts):
    """更新拾取路径显示"""
    if len(pts) > 1:  # 判断拾取路径点数是否足够构成封闭线框
        path_prop.clone(vmi.ccWire(vmi.ccSegments(pts, True)))  # 构造封闭折线段
    else:
        path_prop.clone()  # 否则清空拾取路径显示
    view.updateInTime()  # 刷新视图


def update_plate():
    if len(path_pts) > 1:  # 判断拾取点列是否有效
        cs = view.cameraCS()  # 视角坐标系，cs.axis(2)表示垂直射入屏幕方向

        # 构造初始投影面
        lmax = vmi.pdOBB_Pts(mesh_pd)['size'][0]
        pts = [pt - lmax * cs.axis(2) for pt in path_pts]  # 相对路径点上升足够距离
        top_pd = vmi.pdTriangulate_Polyline(pts)  # 初步三角网格化多边形内部
        top_pd = vmi.pdRemesh(top_pd)  # 网格重划分，使其致密均匀

        # 投影构造下端面
        bottom_pd = vmi.pdRayCast_Pd(top_pd, cs.axis(2), mesh_pd, path_pts[0])  # 投影面上的所有网格点投影到模型表面
        bottom_pd = vmi.pdRemesh(bottom_pd)  # 网格重划分，消除投影造成的拉伸变形
        bottom_pts = vmi.pdBorder_Pts(bottom_pd)  # 获得顺次连接的边界点
        bottom_wire = vmi.ccWire(vmi.ccSegments(bottom_pts, True))  # 构造线，用于放样构造侧面
        bottom_face = vmi.ccSh_Pd(bottom_pd)  # 构造面，用于缝合实体

        # 反向投影构造上端面
        pt_z = {cs.inv().mpt(pt)[2]: pt for pt in bottom_pts}  # 将下端面在视角坐标系中的z坐标构成索引
        zmin, zmax = min(*pt_z), max(*pt_z)  # 计算z坐标的最小值和最大值
        top_pt = pt_z[zmin] - 2 * cs.axis(2)  # 最小值增加一个最小厚度，计算上端面内一点
        top_pts, top_center = [], np.zeros(3)  # 计算上端面边界点和中心
        for pt in bottom_pts:  # 遍历下端面边界点
            pt = vmi.ptOnPlane(pt, top_pt, cs.axis(2))  # 将下端面点投影到上端面
            top_pts.append(pt)  # 收集上端面边界点
            top_center += pt / len(bottom_pts)  # 累加计算上端面中心

        # 构造上端面
        top_wire = vmi.ccWire(vmi.ccSegments(top_pts, True))  # 构造线，用于放样构造侧面
        top_face = vmi.ccFace(top_wire)  # 构造面，用于缝合实体

        # 构造匹配板
        profile = vmi.ccLoft([top_wire, bottom_wire], True, True)  # 放样构造侧面
        plate_solid = vmi.ccSolid(vmi.ccSew([top_face, profile, bottom_face]))  # 缝合构造实体

        # 构造引导孔外圆柱
        hole_outer_radius = 3  # 半径
        hole_outer_height = 10  # 高度
        hole_outer_center = top_center  # 圆台定位到匹配板上端面中心
        hole_outer_solid = vmi.ccCylinder(hole_outer_radius,
                                          hole_outer_height,
                                          hole_outer_center,
                                          -cs.axis(2))  # 构造圆柱体

        # 构造引导孔内圆柱
        hole_inner_radius = 1  # 半径
        hold_depth = zmax - zmin + 2 + hole_outer_height  # 深度，保证切除穿透匹配板
        hole_inner_center = top_center - hole_outer_height * cs.axis(2)  # 圆孔定位到圆台上端面中心
        hole_inner_solid = vmi.ccCylinder(hole_inner_radius,
                                          hold_depth,
                                          hole_inner_center,
                                          cs.axis(2))  # 构造圆柱体

        # 融合匹配板和引导孔构造导板
        plate_solid = vmi.ccBoolean_Union([plate_solid, hole_outer_solid])  # 布尔加匹配板和圆台
        plate_solid = vmi.ccBoolean_Difference([plate_solid, hole_inner_solid])  # 再布尔减圆孔

        plate_prop.clone(plate_solid)  # 更新导板显示
        path_prop.clone(bottom_wire)  # 更新拾取路径
    else:
        plate_prop.clone()  # 拾取路径无效时清空导板显示
    view.updateInTime()  # 刷新视图


if __name__ == '__main__':
    mesh_pd = read_stl()  # 读取STL文件为个性化匹配的目标模型
    if mesh_pd is None:  # 判断用户是否有效读取了模型文件
        vmi.appexit()  # 清理并退出程序

    view = vmi.View()  # 视图
    view.mouse['NoButton']['Wheel'] = Wheel  # 连接鼠标滚轮

    # 面网格模型
    mesh_prop = vmi.PolyActor(view, rgb=[1, 1, 0.6], pickable=True, point=2)
    mesh_prop.mouse['LeftButton']['Press'] = LeftButtonPress  # 连接鼠标左键按下
    mesh_prop.mouse['LeftButton']['PressMove'] = LeftButtonPressMove  # 连接鼠标左键按下移动
    mesh_prop.mouse['LeftButton']['PressMoveRelease'] = LeftButtonPressMoveRelease  # 连接鼠标左键按下移动释放
    mesh_prop.mouse['NoButton']['Wheel'] = Wheel  # 连接鼠标滚轮
    mesh_prop.clone(mesh_pd)  # 更新模型显示

    # 拾取路径点列
    path_pts = []

    # 拾取路径显示
    path_prop = vmi.PolyActor(view, rgb=[1, 0.4, 0.4], line=3)

    # 导板显示
    plate_prop = vmi.PolyActor(view, rgb=[0.4, 0.6, 1])

    view.cameraFit()  # 自动调整视图的视野范围
    vmi.appexec(view)  # 执行主窗口程序
    vmi.appexit()  # 清理并退出程序



import sys

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *
import vtk

import vmi

tr = QObject()
tr = tr.tr


class PolyActor(QObject, vmi.Menu, vmi.Mouse):
    def __init__(self, view: vmi.View, name=None, rgb=None, point=None, line=None, pickable=None):
        QObject.__init__(self)

        self.name = name if name else tr('物体 （Poly)')
        vmi.Menu.__init__(self, name=self.name)

        vmi.Mouse.__init__(self)
        for b in self.mouse:
            for e in self.mouse[b]:
                if e == 'Enter':
                    self.mouse[b][e] = self.mouseEnter
                elif e == 'Leave':
                    self.mouse[b][e] = self.mouseLeave
                elif e == 'Press':
                    self.mouse[b][e] = self.mousePress
                elif e == 'PressMoveRelease':
                    self.mouse[b][e] = (self.mousepass if b == 'RightButton' else self.mouseRelease)
                elif e == 'PressRelease':
                    self.mouse[b][e] = (self.mousepass if b == 'RightButton' else self.mouseRelease)

        self.view = view
        self._Mapper = vtk.vtkPolyDataMapper()
        self._Prop = vtk.vtkActor()
        self._Prop._Prop = self
        self.view._Renderer.AddActor(self._Prop)

        self._Property: vtk.vtkProperty = self._Prop.GetProperty()
        self._RGB, self._Alpha = [1, 1, 1], 1.0
        self.color()
        self._Pickable = False
        self.pickable()
        self._Shade = True
        self.shade(True)
        self._AlwaysOnTop = False
        self.alwaysOnTop(False)

        self._Data = vtk.vtkPolyData()
        self._Bind = self
        self.bind()

        if rgb is not None:
            self.color(rgb=rgb)

        if point is not None:
            self.size(point=point)

        if line is not None:
            self.size(line=line)

        if pickable is not None:
            self.pickable(pickable)

    def __setstate__(self, s):
        self.__init__(s['view'], s['name'])
        self.__dict__.update(s)
        s = self.__dict__

        self._Data.DeepCopy(vmi.loads(s['_Dumps']['_Data']))
        del s['_Dumps']

        for kw in s['_VTKGets']:
            vmi.vtkset(getattr(self, kw), s['_VTKGets'][kw])
        del s['_VTKGets']

        self.bind()

    def __getstate__(self):
        s = self.__dict__.copy()
        for kw in ['menu', 'actions', '__METAOBJECT__']:
            if kw in s:
                del s[kw]

        s['_Dumps'] = {'_Data': vmi.dumps(self._Data)}
        for kw in s['_Dumps']:
            del s[kw]

        s['_VTKGets'] = {name: vmi.vtkget(getattr(self, name)) for name in
                         ['_Mapper', '_Prop', '_Property']}
        for kw in s['_VTKGets']:
            del s[kw]
        return s

    def data(self):
        return self._Bind._Data

    def bind(self, other=None):
        if hasattr(other, '_Data'):
            self._Bind = other
        self._Mapper.SetInputData(self.data())
        self._Prop.SetMapper(self._Mapper)
        self.view.updateInTime()

    def clone(self, other=None):
        if other is None:
            self.data().ShallowCopy(self.data().__class__())
        elif hasattr(other, '_Data'):
            self.data().ShallowCopy(other.data())
        elif isinstance(other, self.data().__class__):
            self.data().ShallowCopy(other)
        elif isinstance(other, vmi.TopoDS_Shape):
            other = vmi.ccPd_Sh(other)
            self.clone(other)
        elif isinstance(other, vmi.Polyhedron_3):
            other = vmi.cgPd_Ph(other)
            self.clone(other)
        else:
            raise TypeError(type(other))

    def alwaysOnTop(self, arg=None):
        self.view.updateInTime()
        if arg is True:
            self._AlwaysOnTop = True
            self._Mapper.SetRelativeCoincidentTopologyLineOffsetParameters(0, -66000)
            self._Mapper.SetRelativeCoincidentTopologyPolygonOffsetParameters(0, -66000)
            self._Mapper.SetRelativeCoincidentTopologyPointOffsetParameter(-66000)
        elif arg is False:
            self._AlwaysOnTop = False
            self._Mapper.SetRelativeCoincidentTopologyLineOffsetParameters(-1, -1)
            self._Mapper.SetRelativeCoincidentTopologyPolygonOffsetParameters(-1, -1)
            self._Mapper.SetRelativeCoincidentTopologyPointOffsetParameter(-1)
        return self._AlwaysOnTop

    def color(self, rgb=None, alpha=None):
        self.view.updateInTime()
        if rgb is not None:
            self._RGB = rgb
        if alpha is not None:
            self._Alpha = alpha
        self._Property.SetColor(self._RGB)
        self._Property.SetOpacity(self._Alpha)
        return {'rgb': self._RGB, 'alpha': self._Alpha}

    def colorLight(self):
        c = QColor()
        c.setRgbF(self._RGB[0], self._RGB[1], self._RGB[2])
        c = c.lighter(120)
        self._Property.SetColor(c.redF(), c.greenF(), c.blueF())
        self.view.updateInTime()

    def colorDark(self):
        c = QColor()
        c.setRgbF(self._RGB[0], self._RGB[1], self._RGB[2])
        c = c.darker(120)
        self._Property.SetColor(c.redF(), c.greenF(), c.blueF())
        self.view.updateInTime()

    def colorNormal(self):
        self._Property.SetColor(self._RGB)
        self.view.updateInTime()

    def delete(self):
        self._Data.ReleaseData()
        self.view._Renderer.RemoveActor(self._Prop)

    def pickable(self, arg=None):
        if arg is True:
            self._Pickable = True
        elif arg is False:
            self._Pickable = False
        self._Prop.SetPickable(1 if self._Pickable else 0)
        self.view.updateInTime()
        return self._Pickable

    def visible(self, visible=None):
        if visible is not None:
            self._Prop.SetVisibility(1 if visible else 0)
            self.view.updateInTime()
        return True if self._Prop.GetVisibility() else False

    def visibleToggle(self, *args):
        self.visible(not self.visible())

    def rep(self, arg=None):
        self.view.updateInTime()
        if isinstance(arg, str):
            if arg.lower() == 'points':
                self._Property.SetRepresentationToPoints()
            elif arg.lower() == 'wireframe':
                self._Property.SetRepresentationToWireframe()
            elif arg.lower() == 'surface':
                self._Property.SetRepresentationToSurface()
            elif arg.lower() == 'toggle':
                r = (self._Property.GetRepresentation() + 1) % 3
                self.rep(('points', 'wireframe', 'surface')[r])
        return ('points', 'wireframe', 'surface')[self._Property.GetRepresentation()]

    def shade(self, arg=None):
        self.view.updateInTime()
        if arg is True:
            self._Shade = True
            self._Property.SetAmbient(0)
            self._Property.SetDiffuse(1)
        elif arg is False:
            self._Shade = False
            self._Property.SetAmbient(1)
            self._Property.SetDiffuse(0)
        return self._Shade

    def size(self, **kwargs):
        self.view.updateInTime()
        if 'point' in kwargs:
            self._Property.SetPointSize(kwargs['point'])
        if 'line' in kwargs:
            self._Property.SetLineWidth(kwargs['line'])
        return {'point': self._Property.GetPointSize(),
                'line': self._Property.GetLineWidth()}

    def mouseEnter(self, **kwargs):
        self.colorLight()
        self.mousepass()

    def mousePress(self, **kwargs):
        self.colorDark()
        self.mousepass()

    def mouseRelease(self, **kwargs):
        self.colorNormal()
        self.mousepass()

    def mouseLeave(self, **kwargs):
        self.colorNormal()
        self.mousepass()


class TextActor(QObject, vmi.Menu, vmi.Mouse):
    def __init__(self, view: vmi.View, name=None):
        QObject.__init__(self)

        self.name = name if name else tr('文字 （Text)')
        vmi.Menu.__init__(self, name=self.name)

        vmi.Mouse.__init__(self)
        for b in self.mouse:
            for e in self.mouse[b]:
                if e == 'Enter':
                    self.mouse[b][e] = self.mouseEnter
                elif e == 'Leave':
                    self.mouse[b][e] = self.mouseLeave
                elif e == 'Press':
                    self.mouse[b][e] = self.mousePress
                elif e == 'PressMoveRelease':
                    self.mouse[b][e] = self.mouseRelease
                elif e == 'PressRelease':
                    self.mouse[b][e] = self.mouseRelease

        self.view = view
        self._Prop = vtk.vtkTextActor()
        self._Prop._Prop = self
        self.view._Renderer.AddActor(self._Prop)

        self._Prop.SetTextScaleModeToNone()

        self._Property: vtk.vtkTextProperty = self._Prop.GetTextProperty()
        self._Property.SetFontSize(0)
        self._Property.SetFontFamily(vtk.VTK_FONT_FILE)
        self._Property.SetLineSpacing(1.2)
        self._Property.SetShadow(0)

        self._Text = ''
        self.text()

        self._Aligns = {'left': vtk.VTK_TEXT_LEFT, 'right': vtk.VTK_TEXT_RIGHT,
                        'top': vtk.VTK_TEXT_TOP, 'bottom': vtk.VTK_TEXT_BOTTOM,
                        'center': vtk.VTK_TEXT_CENTERED, }

        self._Align = ['left', 'bottom']
        self.align()
        self._Pos = [0, 0]
        self.pos()
        self._Font = 'alizh'
        self.font()
        self._FontSize = 16
        self.size()

        self._RGB, self._Alpha = [0.2, 0.2, 0.2], 1.0
        self.color()
        self._Pickable = True
        self.pickable()

    def __setstate__(self, s):
        self.__init__(s['view'], s['name'])
        self.__dict__.update(s)
        s = self.__dict__

        del s['_Dumps']

        for kw in s['_VTKGets']:
            vmi.vtkset(getattr(self, kw), s['_VTKGets'][kw])
        del s['_VTKGets']

    def __getstate__(self):
        s = self.__dict__.copy()
        for kw in ['menu', 'actions', '__METAOBJECT__']:
            if kw in s:
                del s[kw]

        s['_Dumps'] = {}
        for kw in s['_Dumps']:
            del s[kw]

        s['_VTKGets'] = {name: vmi.vtkget(getattr(self, name)) for name in
                         ['_Prop', '_Property']}
        for kw in s['_VTKGets']:
            del s[kw]
        return s

    def text(self, text=None):
        if text is not None:
            self._Text = text
        self._Prop.SetInput(self._Text)
        self.view.updateInTime()

    def align(self, h=None, v=None):
        if h is not None:
            self._Align[0] = h
        if v is not None:
            self._Align[1] = v
        self._Property.SetJustification(self._Aligns[self._Align[0]])
        self._Property.SetVerticalJustification(self._Aligns[self._Align[1]])
        self.view.updateInTime()
        return self._Align

    def pos(self, x=None, y=None):
        if x is not None:
            self._Pos[0] = x
        if y is not None:
            self._Pos[1] = y
        self._Prop.SetPosition(self._Pos)
        self.view.updateInTime()
        return self._Pos

    def font(self, font=None):
        if font is not None and font in vmi.appFonts:
            self._Font = font
        self._Property.SetFontFile(vmi.appFonts[self._Font])
        self.view.updateInTime()
        return self._Font

    def color(self, rgb=None, alpha=None):
        self.view.updateInTime()
        if rgb is not None:
            self._RGB = rgb
        if alpha is not None:
            self._Alpha = alpha
        self._Property.SetColor(self._RGB)
        self._Property.SetOpacity(self._Alpha)
        return {'rgb': self._RGB, 'alpha': self._Alpha}

    def colorLight(self):
        c = QColor()
        c.setRgbF(self._RGB[0], self._RGB[1], self._RGB[2])
        c = c.lighter(120)
        self._Property.SetColor(c.redF(), c.greenF(), c.blueF())
        self.view.updateInTime()

    def colorDark(self):
        c = QColor()
        c.setRgbF(self._RGB[0], self._RGB[1], self._RGB[2])
        c = c.darker(120)
        self._Property.SetColor(c.redF(), c.greenF(), c.blueF())
        self.view.updateInTime()

    def colorNormal(self):
        self._Property.SetColor(self._RGB)
        self.view.updateInTime()

    def delete(self):
        self.view._Renderer.RemoveActor(self._Prop)

    def pickable(self, arg=None):
        if arg is True:
            self._Pickable = True
        elif arg is False:
            self._Pickable = False
        self._Prop.SetPickable(1 if self._Pickable else 0)
        self.view.updateInTime()
        return self._Pickable

    def visible(self, visible=None):
        if visible is not None:
            self._Prop.SetVisibility(1 if visible else 0)
            self.view.updateInTime()
        return True if self._Prop.GetVisibility() else False

    def visibleToggle(self, *args):
        self.visible(not self.visible())

    def size(self, **kwargs):
        if 'font' in kwargs:
            self._FontSize = kwargs['font']
        self._Property.SetFontSize(self._FontSize)
        self.view.updateInTime()
        return self._FontSize

    def mouseEnter(self, **kwargs):
        self.colorLight()
        self.mousepass()

    def mousePress(self, **kwargs):
        self.colorDark()
        self.mousepass()

    def mouseRelease(self, **kwargs):
        self.colorNormal()
        self.mousepass()

    def mouseLeave(self, **kwargs):
        self.colorNormal()
        self.mousepass()


class ImageSlice(QObject, vmi.Menu, vmi.Mouse):
    """场景表示，图像数据的断层表示
    vtk.vtkImageData -> vtk.vtkImageSlice"""

    def __init__(self, view: vmi.View, name=None):
        QObject.__init__(self)

        self.name = name if name else tr('图面 （Image slice)')
        vmi.Menu.__init__(self, name=self.name)

        self.actions = {'Visible': QAction(''),
                        'SlicePlaneValueNormal': QAction(''),
                        'SlicePlaneAxial': QAction(tr('横断位 (Axial)')),
                        'SlicePlaneSagittal': QAction(tr('矢状位 (Sagittal)')),
                        'SlicePlaneCoronal': QAction(tr('冠状位 (Coronal)')),
                        'WindowValue': QAction(''),
                        'WindowAuto': QAction(tr('自动 (Auto)')),
                        'WindowBone': QAction(tr('骨骼 (Bone)')),
                        'WindowSoft': QAction(tr('组织 (Soft)'))}

        self.actions['Visible'].triggered.connect(self.visibleToggle)
        self.actions['SlicePlaneAxial'].triggered.connect(self.slicePlaneAxial)
        self.actions['SlicePlaneSagittal'].triggered.connect(self.slicePlaneSagittal)
        self.actions['SlicePlaneCoronal'].triggered.connect(self.slicePlaneCoronal)
        self.actions['WindowAuto'].triggered.connect(self.windowAuto)
        self.actions['WindowBone'].triggered.connect(self.windowBone)
        self.actions['WindowSoft'].triggered.connect(self.windowSoft)

        def aboutToShow():
            self.actions['Visible'].setText(
                tr('可见 (Visible)') + ' = ' + (tr('显示 (Show)') if self.visible() else tr('隐藏 (Hide)')))
            self.actions['SlicePlaneValueNormal'].setText(
                tr('法向 (Normal)') + ' = ' + repr(self._BindSlicePlane._SlicePlane.GetNormal()))
            self.actions['WindowValue'].setText(
                tr('宽/位 (W/L)') + ' = ' + (repr(tuple(self._Window))))

            self.menu.clear()
            self.menu.addAction(self.actions['Visible'])

            menu = QMenu(tr('切面 (Slice plane)'))
            menu.addAction(self.actions['SlicePlaneValueNormal'])
            menu.addSeparator()
            menu.addAction(self.actions['SlicePlaneAxial'])
            menu.addAction(self.actions['SlicePlaneSagittal'])
            menu.addAction(self.actions['SlicePlaneCoronal'])
            self.menu.addMenu(menu)

            menu = QMenu(tr('窗 (Window)'))
            menu.addAction(self.actions['WindowValue'])
            menu.addSeparator()
            menu.addAction(self.actions['WindowAuto'])
            menu.addAction(self.actions['WindowBone'])
            menu.addAction(self.actions['WindowSoft'])
            self.menu.addMenu(menu)

        self.menu.aboutToShow.connect(aboutToShow)

        vmi.Mouse.__init__(self, menu=self)
        self.mouse['NoButton']['Wheel'] = self.slice
        self.mouse['LeftButton']['PressMove'] = self.windowMove

        self.view = view
        self._Mapper = vtk.vtkImageResliceMapper()
        self._Prop = vtk.vtkImageSlice()
        self._Prop._Prop = self
        self.view._Renderer.AddViewProp(self._Prop)

        self._Property = self._Prop.GetProperty()
        self._Property.SetInterpolationTypeToCubic()
        self._Property.SetUseLookupTableScalarRange(1)

        self._Data = vtk.vtkImageData()
        self._Bind = self

        self._SlicePlane = vtk.vtkPlane()
        self._BindSlicePlane = self

        self._LookupTable = vtk.vtkLookupTable()
        self._BindLookupTable = self

        self._Window = [1000, 400]

        self.bind()
        self.bindSlicePlane()
        self.bindLookupTable()
        self.windowSoft()

    def __setstate__(self, s):
        self.__init__(s['view'], s['name'])
        self.__dict__.update(s)
        s = self.__dict__

        self._Data.DeepCopy(vmi.loads(s['_Dumps']['_Data']))
        self._LookupTable.DeepCopy(vmi.loads(s['_Dumps']['_LookupTable']))
        del s['_Dumps']

        for kw in s['_VTKGets']:
            vmi.vtkset(getattr(self, kw), s['_VTKGets'][kw])
        del s['_VTKGets']

        self._Pickable = True
        self.pickable()

        self.bind()
        self.bindSlicePlane()
        self.bindLookupTable()

    def __getstate__(self):
        s = self.__dict__.copy()
        for kw in ['menu', 'menuWindow', 'menuSlicePlane', 'actions', '__METAOBJECT__']:
            if kw in s:
                del s[kw]

        s['_Dumps'] = {'_Data': vmi.dumps(self._Data),
                       '_LookupTable': vmi.dumps(self._LookupTable)}
        for kw in s['_Dumps']:
            del s[kw]

        s['_VTKGets'] = {name: vmi.vtkget(getattr(self, name)) for name in
                         ['_Mapper', '_Prop', '_Property', '_SlicePlane']}
        for kw in s['_VTKGets']:
            del s[kw]

        return s

    def data(self):
        return self._Bind._Data

    def bind(self, other=None):
        if hasattr(other, '_Data'):
            self._Bind = other
        elif other is not None:
            raise AttributeError(other)
        self._Mapper.SetInputData(self.data())
        self._Prop.SetMapper(self._Mapper)
        self.view.updateInTime()

    def bindSlicePlane(self, other=None):
        if hasattr(other, '_SlicePlane'):
            self._BindSlicePlane = other
        self._Mapper.SetSlicePlane(self._BindSlicePlane._SlicePlane)
        self.view.updateInTime()

    def bindLookupTable(self, other=None):
        if hasattr(other, '_LookupTable'):
            self._BindLookupTable = other
        self._Property.SetLookupTable(self._BindLookupTable._LookupTable)
        self.view.updateInTime()

    def clone(self, other):
        if other is None:
            self.data().DeepCopy(self.data().__class__())
        elif hasattr(other, '_Data'):
            self.data().DeepCopy(other.data())
        elif isinstance(other, self.data().__class__):
            self.data().DeepCopy(other)
        else:
            raise TypeError(type(other))

    def delete(self, *args):
        self._Data.ReleaseData()
        self.view._Renderer.RemoveViewProp(self._Prop)

    def pickable(self, arg=None):
        if arg is True:
            self._Pickable = True
        elif arg is False:
            self._Pickable = False
        self._Prop.SetPickable(1 if self._Pickable else 0)
        self.view.updateInTime()
        return self._Pickable

    def visible(self, visible=None):
        if visible is not None:
            self._Prop.SetVisibility(1 if visible else 0)
            self.view.updateInTime()
        return True if self._Prop.GetVisibility() else False

    def visibleToggle(self, *args):
        self.visible(not self.visible())

    def slicePlane(self, face=None, origin=None):
        if isinstance(origin, str):
            if origin.lower() == 'center':
                o = self.data().GetCenter()
                self._BindSlicePlane._SlicePlane.SetOrigin(o[0], o[1], o[2])
            elif origin.lower() in ('b0', 'b1', 'b2', 'b3', 'b4', 'b5'):
                i = int(origin[-1])
                o = list(self.data().GetCenter())
                o[int(i / 2)] = self.data().GetBounds()[i]
                self._BindSlicePlane._SlicePlane.SetOrigin(o[0], o[1], o[2])
        elif origin is not None:
            self._BindSlicePlane._SlicePlane.SetOrigin(origin[0], origin[1], origin[2])
        if isinstance(face, str):
            if face.lower() == 'sagittal':
                self._BindSlicePlane._SlicePlane.SetNormal(1, 0, 0)
            elif face.lower() == 'coronal':
                self._BindSlicePlane._SlicePlane.SetNormal(0, 1, 0)
            elif face.lower() == 'axial':
                self._BindSlicePlane._SlicePlane.SetNormal(0, 0, 1)
        elif face is not None:
            self._BindSlicePlane._SlicePlane.SetNormal(face[0], face[1], face[2])
        self.view.updateInTime()

    def slicePlaneAxial(self):
        self.slicePlane(face='axial')

    def slicePlaneSagittal(self):
        self.slicePlane(face='sagittal')

    def slicePlaneCoronal(self):
        self.slicePlane(face='coronal')

    def window(self, preset=None, width=None, level=None):
        if preset is None:
            if width is not None:
                self._Window[0] = width
            if level is not None:
                self._Window[1] = level

            self._Window[0] = int(self._Window[0])
            self._Window[1] = int(self._Window[1])

            r = [self._Window[1] - 0.5 * self._Window[0],
                 self._Window[1] + 0.5 * self._Window[0]]

            t = self._BindLookupTable._LookupTable
            t.SetNumberOfTableValues(self._Window[0])
            t.SetTableRange(r)
            t.SetBelowRangeColor(0, 0, 0, 1.0)
            t.SetAboveRangeColor(1, 1, 1, 1.0)
            t.SetUseBelowRangeColor(1)
            t.SetUseAboveRangeColor(1)

            for i in range(self._Window[0]):
                v = i / self._Window[0]
                t.SetTableValue(i, (v, v, v, 1.0))
            self.view.updateInTime()

        elif preset.lower() == 'auto':
            r = vtk.vtkImageHistogramStatistics()
            r.SetInputData(self.data())
            r.SetAutoRangePercentiles(1, 99)
            r.SetGenerateHistogramImage(0)
            r.Update()
            r = r.GetAutoRange()

            self.window(width=r[1] - r[0], level=0.5 * (r[0] + r[1]))
        elif preset.lower() == 'bone':
            self.window(width=1000, level=400)
        elif preset.lower() == 'soft':
            self.window(width=350, level=50)

    def windowAuto(self, **kwargs):
        self.window('auto')

    def windowBone(self, **kwargs):
        self.window('bone')

    def windowSoft(self, **kwargs):
        self.window('soft')

    def slice(self, **kwargs):
        o = list(self._BindSlicePlane._SlicePlane.GetOrigin())
        n = self._BindSlicePlane._SlicePlane.GetNormal()
        dim = self.data().GetDimensions()
        dxyz = self.data().GetSpacing()

        n = [abs(_) / (n[0] ** 2 + n[1] ** 2 + n[2] ** 2) ** 0.5 for _ in n]
        dn = n[0] * dxyz[0] + n[1] * dxyz[1] + n[2] * dxyz[2]

        for i in range(3):
            if dim[i] > 1:
                o[i] += kwargs['delta'] * dn * n[i]

        if vtk.vtkMath.PlaneIntersectsAABB(self.data().GetBounds(), n, o) == 0:
            self.slicePlane(origin=o)

    def windowMove(self, **kwargs):
        dx, dy = self.view.mouseOverDisplay()

        r = self.data().GetScalarRange()
        t = (r[1] - r[0]) / 2048
        self._Window[0] += t * dx
        self._Window[1] -= t * dy

        self._Window[0] = max(self._Window[0], 0)
        self._Window[0] = min(self._Window[0], r[1] - r[0])
        self._Window[1] = max(self._Window[1], r[0])
        self._Window[1] = min(self._Window[1], r[1])
        self.window()


class ImageVolume(QObject, vmi.Menu, vmi.Mouse):
    """场景表示，图像数据的立体表示
    vtk.vtkImageData -> vtk.vtkVolume"""

    def __init__(self, view: vmi.View, name=None):
        QObject.__init__(self)

        self.name = name if name else tr('图体 （Image volume)')
        vmi.Menu.__init__(self, name=self.name)

        self.actions = {'Visible': QAction(''),
                        'Threshold': QAction(tr('阈值 (Threshold)'))}

        self.actions['Visible'].triggered.connect(self.visibleToggle)
        self.actions['Threshold'].triggered.connect(self.threshold)

        def aboutToShow():
            self.actions['Visible'].setText(
                tr('可见 (Visible)') + ' = ' + (tr('显示 (Show)') if self.visible() else tr('隐藏 (Hide)')))

            self.menu.clear()
            self.menu.addAction(self.actions['Visible'])

            menu = QMenu(tr('风格 (Style)'))
            menu.addAction(self.actions['Threshold'])
            self.menu.addMenu(menu)

        self.menu.aboutToShow.connect(aboutToShow)

        vmi.Mouse.__init__(self)

        self.view = view
        self._Mapper = vtk.vtkGPUVolumeRayCastMapper()
        self._Prop = vtk.vtkVolume()
        self._Prop._Prop = self
        self.view._Renderer.AddVolume(self._Prop)

        self._Mapper.SetBlendModeToComposite()
        self._Mapper.SetMaxMemoryInBytes(4096)
        self._Mapper.SetMaxMemoryFraction(1)
        self._Mapper.SetAutoAdjustSampleDistances(0)
        self._Mapper.SetLockSampleDistanceToInputSpacing(1)
        self._Mapper.SetUseJittering(1)

        self._Property = self._Prop.GetProperty()
        self._Property.SetInterpolationTypeToLinear()
        self._Property.SetAmbient(0.1)
        self._Property.SetDiffuse(0.9)
        self._Property.SetShade(1)

        self._Data = vtk.vtkImageData()
        self._Bind = self

        self._Color = {0: [1, 1, 1]}
        self._ScalarOpacity = {0: 0, 400: 1}
        self._GradientOpacity = {}

        self._Pickable = True
        self.pickable()

        self.bind()
        self.color()
        self.opacityScalar()
        self.opacityGradient()

    def __setstate__(self, s):
        self.__init__(s['view'], s['name'])
        self.__dict__.update(s)
        s = self.__dict__

        self._Data.DeepCopy(vmi.loads(s['_Dumps']['_Data']))
        del s['_Dumps']

        for kw in s['_VTKGets']:
            vmi.vtkset(getattr(self, kw), s['_VTKGets'][kw])
        del s['_VTKGets']

        self.bind()
        self.color()
        self.opacityScalar()
        self.opacityGradient()

    def __getstate__(self):
        s = self.__dict__.copy()
        for kw in ['menu', 'actions', '__METAOBJECT__']:
            if kw in s:
                del s[kw]

        s['_Dumps'] = {'_Data': vmi.dumps(self._Data),
                       '_Color': self._Color}
        for kw in s['_Dumps']:
            del s[kw]

        s['_VTKGets'] = {name: vmi.vtkget(getattr(self, name)) for name in
                         ['_Mapper', '_Prop', '_Property']}
        for kw in s['_VTKGets']:
            del s[kw]

        return s

    def data(self):
        return self._Bind._Data

    def bind(self, other=None):
        if hasattr(other, '_Data'):
            self._Bind = other
        elif other is not None:
            raise AttributeError(other)
        self._Mapper.SetInputData(self.data())
        self._Prop.SetMapper(self._Mapper)
        self.view.updateInTime()

    def clone(self, other):
        if other is None:
            self.data().ShallowCopy(self.data().__class__())
        elif hasattr(other, 'data'):
            self.data().ShallowCopy(other.data())
        elif isinstance(other, self.data().__class__):
            self.data().ShallowCopy(other)
        else:
            raise TypeError(type(other))

    def delete(self):
        # self._Data.ReleaseData()
        image = vtk.vtkImageData()
        image.SetExtent(0, 1, 0, 0, 0, 0)
        image.AllocateScalars(vtk.VTK_SHORT, 1)
        self.clone(image)
        self.view._Renderer.RemoveVolume(self._Prop)

    def pickable(self, arg=None):
        if arg is True:
            self._Pickable = True
        elif arg is False:
            self._Pickable = False
        self._Prop.SetPickable(1 if self._Pickable else 0)
        self.view.updateInTime()
        return self._Pickable

    def visible(self, visible=None):
        if visible is not None:
            self._Prop.SetVisibility(1 if visible else 0)
            self.view.updateInTime()
        return True if self._Prop.GetVisibility() else False

    def visibleToggle(self, *args):
        self.visible(not self.visible())

    def color(self, scalar_rgb=None):
        if scalar_rgb is not None:
            self._Color = scalar_rgb
            return self.color()
        else:
            f = vtk.vtkColorTransferFunction()
            for x in self._Color:
                r, g, b = self._Color[x]
                f.AddRGBPoint(x, r, g, b)

            self._Property.SetColor(f)
            return self._Color

    def opacityScalar(self, scalar_opacity=None):
        if scalar_opacity is not None:
            self._ScalarOpacity = scalar_opacity
            return self.opacityScalar()
        else:
            f = vtk.vtkPiecewiseFunction()
            for x in self._ScalarOpacity:
                f.AddPoint(x, self._ScalarOpacity[x])
            self._Property.SetScalarOpacity(f)
            return self._Property

    def opacityGradient(self, gradient_opacity=None):
        if gradient_opacity is not None:
            self._GradientOpacity = gradient_opacity
            return self.opacityScalar()
        else:
            f = vtk.vtkPiecewiseFunction()
            for x in self._GradientOpacity:
                f.AddPoint(x, self._GradientOpacity[x])
            self._Property.SetGradientOpacity(f)
            return self._Property

    def threshold(self, *args, value=None):
        if value is None:
            value = vmi.askInt(vtk.VTK_SHORT_MIN, 0, vtk.VTK_SHORT_MAX, tr('阈值 (Threshold value)'))
        if value is not None:
            self.opacityScalar({value - 1: 0, value: 1})



import locale
import pathlib
import shutil
import tempfile

import vtk

from OCC.Core.TopoDS import *
from OCC.Core.STEPControl import *
from OCC.Core.IFSelect import *

vtkget_ignore = ('GetDebug', 'GetGlobalReleaseDataFlag', 'GetGlobalWarningDisplay', 'GetReferenceCount',
                 'GetAAFrames', 'GetFDFrames', 'GetSubFrames', 'GetUseConstantFDOffsets', 'GetStereoCapableWindow',
                 'GetForceCompileOnly', 'GetGlobalImmediateModeRendering', 'GetImmediateModeRendering',
                 'GetScalarMaterialMode', 'GetReleaseDataFlag')


def vtkset(self, gets: dict):
    for getname in gets:
        setname = getname.replace('Get', 'Set', 1)
        try:
            getattr(self, setname)(gets[getname])
        except TypeError:
            pass


def vtkget(self):
    gets = {}
    for getname in dir(self):
        if getname.startswith('Get'):
            setname = getname.replace('Get', 'Set', 1)
            if hasattr(self, setname) and getname not in vtkget_ignore:
                try:
                    a = getattr(self, getname)()
                    if 'vtk' not in str(type(a)):
                        gets[getname] = a
                except TypeError:
                    pass
    return gets


def dumps(arg):
    if isinstance(arg, vtk.vtkImageData):
        if arg.GetNumberOfPoints() == 0:
            return ['vtkImageData', bytes()]
        else:
            with tempfile.TemporaryDirectory() as p:
                p = pathlib.Path(p) / '.nii'
                w = vtk.vtkNIFTIImageWriter()
                w.SetFileName(str(p))
                w.SetInputData(arg)
                w.Update()
                return ['vtkImageData', p.read_bytes()]
    elif isinstance(arg, vtk.vtkPolyData):
        if arg.GetNumberOfPoints() == 0:
            return ['vtkPolyData', bytes()]
        else:
            with tempfile.TemporaryDirectory() as p:
                p = pathlib.Path(p) / '.vtp'
                w = vtk.vtkPolyDataWriter()
                w.SetFileTypeToBinary()
                w.SetFileName(str(p))
                w.SetInputData(arg)
                w.Update()
                return ['vtkPolyData', p.read_bytes()]
    elif isinstance(arg, vtk.vtkLookupTable):
        p = {'GetNumberOfTableValues': arg.GetNumberOfTableValues(),
             'GetTableRange': arg.GetTableRange(),
             'GetBelowRangeColor': arg.GetBelowRangeColor(),
             'GetAboveRangeColor': arg.GetAboveRangeColor(),
             'GetUseBelowRangeColor': arg.GetUseBelowRangeColor(),
             'GetUseAboveRangeColor': arg.GetUseAboveRangeColor(),
             'GetTableValue': []}
        for i in range(arg.GetNumberOfTableValues()):
            p['GetTableValue'].append(arg.GetTableValue(i))
        return ['vtkLookupTable', p]
    elif isinstance(arg, TopoDS_Shape):
        if arg.IsNull():
            return ['TopoDS_Shape', bytes()]
        else:
            with tempfile.TemporaryDirectory() as p:
                p = pathlib.Path(p) / '.stp'
                w = STEPControl_Writer()
                w.Transfer(arg, STEPControl_AsIs)
                status = w.Write(str(p))
                if status == IFSelect_RetDone:
                    return ['TopoDS_Shape', p.read_bytes()]
                raise Exception(('IFSelect_RetVoid', 'IFSelect_RetDone', 'IFSelect_RetError',
                                 'IFSelect_RetFail', 'IFSelect_RetStop')[status])
    raise TypeError(type(arg))


def loads(arg):
    if arg[0] == 'vtkImageData':
        if len(arg[1]) == 0:
            return vtk.vtkImageData()
        else:
            with tempfile.TemporaryDirectory() as p:
                p = pathlib.Path(p) / '.nii'
                p.write_bytes(arg[1])
                r = vtk.vtkNIFTIImageReader()
                r.SetFileName(str(p))
                r.Update()
                return r.GetOutput()
    elif arg[0] == 'vtkPolyData':
        if len(arg[1]) == 0:
            return vtk.vtkPolyData()
        else:
            with tempfile.TemporaryDirectory() as p:
                p = pathlib.Path(p) / '.vtp'
                p.write_bytes(arg[1])
                r = vtk.vtkPolyDataReader()
                r.SetFileName(str(p))
                r.Update()
                return r.GetOutput()
    elif arg[0] == 'vtkLookupTable':
        r = vtk.vtkLookupTable()
        r.SetNumberOfTableValues(arg[1]['GetNumberOfTableValues'])
        r.SetTableRange(arg[1]['GetTableRange'])
        r.SetBelowRangeColor(arg[1]['GetBelowRangeColor'])
        r.SetAboveRangeColor(arg[1]['GetAboveRangeColor'])
        r.SetUseBelowRangeColor(arg[1]['GetUseBelowRangeColor'])
        r.SetUseAboveRangeColor(arg[1]['GetUseAboveRangeColor'])
        for i in range(arg[1]['GetNumberOfTableValues']):
            r.SetTableValue(i, arg[1]['GetTableValue'][i])
        return r
    elif arg[0] == 'TopoDS_Shape':
        if len(arg[1]) == 0:
            return TopoDS_Shape()
        else:
            with tempfile.TemporaryDirectory() as p:
                p = pathlib.Path(p) / '.stp'
                p.write_bytes(arg[1])
                r = STEPControl_Reader()
                status = r.ReadFile(str(p))
                if status == IFSelect_RetDone:
                    r.TransferRoots()
                    return r.OneShape()
                else:
                    raise Exception(('IFSelect_RetVoid', 'IFSelect_RetDone', 'IFSelect_RetError',
                                     'IFSelect_RetFail', 'IFSelect_RetStop')[status])
    raise TypeError(arg[0])


def vtkread_STL(file_name):
    with tempfile.TemporaryDirectory() as p:
        p = pathlib.Path(p) / '.stl'
        shutil.copyfile(file_name, p)
        r = vtk.vtkSTLReader()
        r.SetFileName(str(p))
        r.Update()
        return r.GetOutput()


def vtkwrite_STL(file_name, file_type={'ASCII': vtk.VTK_ASCII,
                                       'Binary': vtk.VTK_BINARY}['Binary']):
    with tempfile.TemporaryDirectory() as p:
        p = pathlib.Path(p) / '.stl'
        w = vtk.vtkSTLWriter()
        w.SetFileName(str(p))
        w.SetFileType(file_type)
        w.Update()
        shutil.copyfile(p, file_name)


def diffEncode(text):
    return text.encode() != text.encode(locale.getdefaultlocale()[1])


if __name__ == '__main__':
    reader = vtk.vtkSTLReader()
    reader.SetFileName('C:/Users/Medraw/Desktop/1.stl')
    reader.Update()

    loads(dumps(reader.GetOutput()))

    from time import sleep

    sleep(5)
    print('EXIT_SUCCESS')

    # with shelve.open('db', 'n', HIGHEST_PROTOCOL) as db:
    #     db['w'] = tempfile.TemporaryFile( )
